#include <iostream>
#include <iomanip>
#include <fstream>
#include <math.h>
#include <stdio.h>
#include <cstdlib>

using namespace std;

// definition of Cell class
class Cell{
public:
	int dir;	// flow direction
				// NN --> 1, NE --> 2, EE --> 3, SE --> 4, SS --> 5, SW-->6, WW-->7, NW-->,8
				// river mouth --> 0, sea or pit --> -1
	int basin; 	// upstream grid number
	int riverNumber;			// catchment number ԍ
	class Cell* upStreamGrid[7]; // upstream grid 㗬Obh
	class Cell* downStreamGrid;  // lowerstream grid Obh

	//ZʐόvZp
	float latitude;				// ܓx [b] latitude [second]
	float longitude;			// ox [b] longitude [second]
	int topoScale;				// 3sԕ\ɑ΂䗦 ratio of spatial resolution to 3 second (10 means 30 second)
	float longitudinalLength;	/* ox30b̒[m] */
	float latitudinalLength;	/* ܓx30b̒[m] */

	void setInit(class Cell** cell2, int i, int j);
	void setDownStreamGrid(int row, int col, class Cell** cell2);
	void setUpStreamGrid();
	void setRiverNumber();
	double cellArea(int rowStep, int colStep, int row, int col, float SWlat, float SWlon, int topoScale);
};


int main()
{
//	errno_t error;

	//read study area for cliping topographic data.
	cout << "Reading location of study area." << endl;
	ifstream finClipArea("./clipArea.dat");
	if(!finClipArea){
		cout << "clipArea.dat cannot open." << endl;
		cout << endl;
		cout << "Program ends. Hit any key!" << endl;
		char ch;
		cin.get(ch);
		return 1;
	}

	int wCol, wRow;
	double wSWLON, wSWLAT;

	int wTopoScale;

	finClipArea >> wTopoScale;
	cout << "  Topography scale : " << wTopoScale << endl;
	const int TopoScale = wTopoScale;	// HydroSHEDSn`f[^̍ō\3bɑ΂gpf[^̔䗦 Topography scale to 3 second (finest resolution of HydroSHEDS)
	double const Dx = 0.000833333*(double)TopoScale; // cellsize in degrees

	finClipArea >> wCol; cout << "  Number of column : " << wCol << endl;
	finClipArea >> wRow; cout << "  Number of row : " << wRow << endl;
	// Lower Left corner longitude (deg.)
	finClipArea >> wSWLON; cout << "  Longitude (degree) of South West corner : " << wSWLON << endl;
	// Lower left corner Latitude (deg.)
	finClipArea >> wSWLAT; cout << "  Latitude (degree) of South West corner : " << wSWLAT << endl;

	const int Col = wCol;     // column number of study area
	const int Row = wRow;     // row number of study area
	const double SWLON = wSWLON; // South west corner of longitude (degree)
	const double SWLAT = wSWLAT; // South west corner of latiitude (degree)

	finClipArea.close();


	//--- Memory allocation ----------------------------------------------------
	class Cell** cellObject;

	cout << "  Memory allocation." << endl;
	cellObject = new class Cell*[Row];
	for(int i=0; i<Row; i++) {
		cellObject[i] = new class Cell[Col];
	}

	// ---Initial setting ------------------------------------------------------
	for(int i=0; i < Row; i++){
		for(int j=0; j < Col; j++){
			cellObject[i][j].setInit(cellObject, i, j);
		}	
	}

	//---------read flow direction data----------------------------------
	// set flow direction, accumulated area and river number information 
	FILE *fDown;
	FILE *fBasin;
	//if(error = fopen_s(&fDown, "./input/flowDir.bin", "rb") != NULL) {
	if ((fDown = fopen("./input/flowDir.bin", "rb")) == NULL) {
		cout << "down.bin cannot open." << endl;
		exit(1);
	}

	//if(error = fopen_s(&fBasin, "./input/flowAcc.bin", "rb") != NULL) {
	if ((fBasin = fopen("./input/flowAcc.bin", "rb")) == NULL) {
		cout << "flowAcc.bin cannot open." << endl;
		exit(1);
	}

	for(int i=Row-1; i >= 0; i--) { // Topo data arranges from south to north
		for(int j=0; j < Col; j++){
			fread(&cellObject[i][j].dir, sizeof(int), 1, fDown);
			fread(&cellObject[i][j].basin, sizeof(int), 1, fBasin);

			if(cellObject[i][j].dir < 0) {
				continue;
			}
			cellObject[i][j].setDownStreamGrid(i, j, cellObject);
		}
	}
	fclose(fDown);

	//-----------Set upstream grid-----------------------------------------------
	cout << "  Setting upstream grid number.\n";

	for(int i=0; i < Row; i++){
		for(int j=0; j < Col; j++){
			if(cellObject[i][j].dir < 0) { // skip at the boundary and sea grid 
				continue;
			}
			if(cellObject[i][j].downStreamGrid == NULL) {
				continue;
			}
			cellObject[i][j].setUpStreamGrid();
		}
	}

	//---------- Set outlet ---------------------------------------------
	// setting column and row number
	cout << "Input column number to extract data!" << endl;
	cout << "Column number begins from 1 from east." << endl;
	cin >> wCol;

	string ss;
	if (wCol > Col || wCol < 0) {
		cout << "Column number is out of range!" << endl;
		return 1;
	}

	cout << "Input row number to extract data!" << endl;
	cout << "Row number begins from 1 from south." << endl;
	cin >> wRow;
	if (wRow > Row || wRow < 0) {
		cout << "Row number is out of range!" << endl;
		return 1;
	}
	
	wRow = Row - wRow + 1; // change row from north

	cout << "  Setting the outlet.\n";

	if (cellObject[wRow-1][wCol-1].dir < 0) {
		cout << "Out of range!" << endl;
		return 1;
	}

	int rNum = 2; // river number NOT for study area

	for(int i=0; i < Row; i++){
		for(int j=0; j < Col; j++){
			if(cellObject[i][j].dir == 0) { // river mouth
				// setting river number at mouth
				cellObject[i][j].riverNumber = rNum;
			}
		}
	}

	// river number for study area
	cellObject[wRow-1][wCol-1].riverNumber = 1; 

	//---------------extract basin---------------------------------------------------
	cout << "  Extracting river basin.\n";

	for (int i=0; i < Row; i++) {
		for (int j=0; j < Col; j++) {
			if (cellObject[i][j].basin == 1) { // most upstream grid
				cellObject[i][j].setRiverNumber();
			}
		}
	}


	/*--------------------------------writing output files---------------------------------------------*/
	cout <<"Writing output files" << endl;
	char filename[256];
	FILE *fRiverNumber;

	//sprintf_s(filename, 256, "./output/extractBasin.bin", "");
	sprintf(filename, "./output/extractBasin.bin");
	//if(error = fopen_s(&fRiverNumber, filename, "wb") != NULL) {
	if((fRiverNumber = fopen(filename, "wb")) == NULL) {
		cout << "extractBasin.bin file cannot open." << endl;
		exit(1);
	} 

	int RNODATA = -1;
	for(int i=Row-1; i >= 0; i--){ // write from South to North
		for(int j=0; j < Col; j++){
			if (i==0 || i==Row-1 || j==0 || j==Col-1) {
				fwrite(&RNODATA, sizeof(int), 1, fRiverNumber); // write NODATA
			}
			else {
				fwrite(&cellObject[i][j].riverNumber, sizeof(int), 1, fRiverNumber); // write 4 bite integer value
			}
		}
	}
	cout << "  Writing " << filename << " done." <<endl;
	fclose(fRiverNumber);

	// ArcGIS
	//sprintf_s(filename, 256, "./output/extractBasin.asc", "");
	sprintf(filename, "./output/extractBasin.asc");
	ofstream fout4(filename);
	if(!fout4){
		cout << "ESRI ExtractBasin.asc file can not open\n";
		return 1;
	}
	fout4 << "ncols " << Col << "\n";
	fout4 << "nrows " << Row << "\n";
	fout4 << "xllcorner " << SWLON << "\n";
	fout4 << "yllcorner " << SWLAT << "\n";
	fout4 << "cellsize " << Dx << "\n";
	fout4 << "NODATA_value "<< RNODATA;
	
	for(int i=0; i < Row; i++){
		for(int j=0; j < Col; j++){
			if(j == 0) {
				fout4 << "\n";
			}
			if (i==0 || i==Row-1 || j==0 || j==Col-1) {
				fout4 << " " << RNODATA;
			}
			else {
				fout4 << " " << cellObject[i][j].riverNumber;
			}
		}
	}
	cout << "  Writing " << filename << " done." <<endl;
	fout4.close();

	//ʐςvZ.
	double area = 0.0;
	for(int m = 0; m < Row; m++){
		for(int n = 0; n < Col; n++){     // longitude direction from West
			// extractBasinŎw肵ɑȂObh̖ʐς
			if(cellObject[m][n].riverNumber != 1){
				continue;
			}
			area += cellObject[m][n].cellArea(m, n, Row, Col, SWLAT, SWLON, TopoScale);//[m^2]
		}
	}
	area *= 1.0e-6;//[m^2]-->[km^2]
	cout << "Basin area : " << area << " [km^2]" << endl;

	delete[] cellObject;

	cout << endl;
	cout << "Program ends!" << endl;
	// cout << "Program ends. Hit any key!" << endl;
	// char ch;
	// cin.get(ch);
	
	return 0;
}

//========== member functions definition ==================

/*  Initialize flow direction, adjesant grid number and lower grid number */
void Cell::setInit(class Cell** cell2, int i, int j)
{
	riverNumber = -1;
	downStreamGrid=NULL;

	for(int i=0; i < 7; i++){
		upStreamGrid[i] = NULL;
	}
}

/* o"downStreamGrid"ɒl */
void Cell::setDownStreamGrid(int row, int col, class Cell** cell2)
{
	switch(dir){
	case 1:	downStreamGrid = &cell2[row-1][col];
		break;
		
	case 2:	downStreamGrid = &cell2[row-1][col+1];
		break;
		
	case 3:	downStreamGrid = &cell2[row][col+1];
		break;
		
	case 4: downStreamGrid = &cell2[row+1][col+1];
		break;
		
	case 5: downStreamGrid = &cell2[row+1][col];
		break;
		
	case 6: downStreamGrid = &cell2[row+1][col-1];
		break;
		
	case 7: downStreamGrid = &cell2[row][col-1];
		break;
		
	case 8: downStreamGrid = &cell2[row-1][col-1];
		break;
		
	default: break;
	}
}

// setting upstream grid
void Cell::setUpStreamGrid()
{
	for(int i=0; i < 7; i++) {
		if (downStreamGrid->upStreamGrid[i] == NULL){ 
			downStreamGrid->upStreamGrid[i] = this;
			break;
		}
	}
}

// set the river number 
void Cell::setRiverNumber()
{
	int rNum;
	class Cell* object;
	
	object = this;
	while (object->riverNumber == -1) {
		if(object->downStreamGrid == NULL) {
			break;
		}
		object = object->downStreamGrid;
	}
	rNum = object->riverNumber;

	object = this;
	while (object->riverNumber == -1) {
		if (object->downStreamGrid == NULL) {
			break;
		}
		object->riverNumber = rNum;
		object = object->downStreamGrid;
	}
}

double Cell::cellArea (
	int rowStep, // row counter
	int colStep, // column counter
	int nRow, // number of row for the study area
	int nCol, //number of column for the study area
	float SWlat, // South west corner of latitude
	float SWlon, // South west corner of longitude
	int TopoScale // ratio of spatial resolution to 3 second (10 means 30 second)
)
{
	topoScale = TopoScale;

	// setting latitude in 'sec' unit
	latitude  = (SWlat*3600) + 3.0*topoScale*(nRow-rowStep-1) + 1.5*topoScale;

	// setting longitude in 'sec' unit
	longitude = (SWlon*3600) + 3.0*topoScale*colStep + 1.5*topoScale;

	// setting length (km) for 1 degree in longitude direction
	longitudinalLength = (float)(92.6886 * topoScale)*cos(latitude/3600.0/180.0*3.14);

	// setting length (km) for 1 degree in longitude direction
	latitudinalLength = (float)(92.6886 * topoScale);
	
	// setting basin area of cell (m2)
	return (double)longitudinalLength * (double)latitudinalLength;
}

